from dataclasses import dataclass
from math import pi, sin, cos, sqrt
import random


@dataclass
class Position:
    x: float
    y: float

    def distance(self, other) -> float:
        return ((self.x - other.x) ** 2 + (self.y - other.y) ** 2) ** 0.5

    def as_tuple(self) -> tuple:
        return self.x, self.y


def get_random_position(radius: float) -> Position:
    """Generate a random position somewhere within a circle.

    :param radius: The radius within which to generate the random position
    :return: The random position object.
    """
    theta = 2 * pi * random.random()
    r = radius * sqrt(random.random())
    x = r * cos(theta)
    y = r * sin(theta)
    return Position(x, y)


def get_random_position_nearby(radius: float, anchor_pos: Position, anchor_radius: float) -> Position:
    """Generate a random position within range of an anchor position.

    :param radius: The radius within which to generate the random position
    :param anchor_pos: The position to which the generated position should appear near.
    :param anchor_radius: The maximum range the random position can be from the anchor.
    :return: A random position.
    """
    x, y = radius, radius
    while x**2 + y**2 > radius**2:
        theta = 2 * pi * random.random()
        r = anchor_radius * sqrt(random.random())
        x = anchor_pos.x + r * cos(theta)
        y = anchor_pos.y + r * sin(theta)
    return Position(x, y)



"""A simple example of how to use this environment.

Each step, GymD2D returns a dict of {agent_id: agent_observation}
    and requires a dict of {agent_id: actions}.
Agent rewards are also returned in a dict of {agent_id: agent_reward}.

The environment follows the Gym format of providing a game_over/terminal bool,
    but this will always be True as the environment is not episodic.
"""

import gym
import gym_d2d


env = gym.make('D2DEnv-v0')

due_agent = 'DUEAgent()'
obses = env.reset()
game_over = False
for _ in range(10):
    actions_dict = {}
    for agent_id, obs in obses.items():
        if agent_id.startswith('due'):
            action = env.action_space['due'].sample()
            # or action = due_agent.act(obs)
        elif agent_id.startswith('cue'):
            action = env.action_space['cue'].sample()
        else:
            action = env.action_space['mbs'].sample()
        actions_dict[agent_id] = action

    obses, rewards_dict, game_over, info = env.step(actions_dict)
    # env.render()
    print(obses)



from .conversion import dB_to_linear, dBm_to_W
from .id import Id
from .position import Position
from .utils import merge_dicts


# https://en.wikipedia.org/wiki/Johnson%E2%80%93Nyquist_noise#Noise_power_in_decibels
THERMAL_NOISE_POWER_dBm = -121.45  # 1x 180kHz LTE RB
THERMAL_NOISE_POWER_mW = dB_to_linear(THERMAL_NOISE_POWER_dBm)
THERMAL_NOISE_POWER_W = dBm_to_W(THERMAL_NOISE_POWER_dBm)

DEFAULT_DEVICE_CONFIG = {
    'num_PRB': 1,
    'num_subcarriers': 12,
    'subcarrier_spacing_kHz': 15.0,
}
DEFAULT_BASE_STATION_CONFIG = merge_dicts(dict(DEFAULT_DEVICE_CONFIG), {
    'max_tx_power_dBm': 46.0,
    # 'antenna_height_m': 10.0,
    'antenna_height_m': 23.0,
    'tx_antenna_gain_dBi': 17.5,
    'rx_antenna_gain_dBi': 17.5,
    'thermal_noise_dBm': -118.4,
    'noise_figure_dB': 2.0,
    'sinr_dB': -7.0,
    'ix_margin_dB': 2.0,  # accounts for the increase in noise from surrounding cells
    'cable_loss_dB': 2.0,
    'masthead_amplifier_gain_dB': 2.0,
})
DEFAULT_UE_CONFIG = merge_dicts(dict(DEFAULT_DEVICE_CONFIG), {
    'max_tx_power_dBm': 23.0,
    'antenna_height_m': 1.5,
    'tx_antenna_gain_dBi': 0.0,
    'rx_antenna_gain_dBi': 0.0,
    'thermal_noise_dBm': -104.5,
    'noise_figure_dB': 7.0,
    'sinr_dB': -10.0,
    'ix_margin_dB': 3.0,  # accounts for the increase in noise from surrounding cells
    'control_channel_overhead_dB': 1.0,
    'body_loss_dB': 3.0,  # loss from user's body blocking reception
})


class Device:
    def __init__(self, id_, config: dict) -> None:
        super().__init__()
        self.id = Id(id_)
        self.config: dict = config
        self.position: Position = Position(0, 0)

    def eirp_dBm(self, tx_pwr_dBm: float) -> float:
        """Calculate the Effective Isotropically Radiated Power (EIRP).

        How much power (in dBm) the device radiates when considering factors such as antenna gain.

        :param tx_pwr_dBm: The transmission power in dBm.
        :returns: The EIRP in dBm.
        """
        # - (10 * log10(self._subcarrier_quantity)) \
        return tx_pwr_dBm + self.tx_antenna_gain_dBi - self.ix_margin_dB

    def rx_signal_level_dBm(self, eirp_dBm: float, path_loss_dB: float) -> float:
        """Calculate the received signal strength.

        :param eirp_dBm: The transmitted EIRP in dBm.
        :param path_loss_dB: The path loss that has occurred in dB.
        :returns: The received signal level in dBm.
        """
        # - self._thermal_noise_dBm \
        # - self._noise_figure_dB \
        # - self._sinr_dB
        return eirp_dBm - path_loss_dB + self.rx_antenna_gain_dBi

    @property
    def rx_sensitivity_dBm(self) -> float:
        return self.rx_noise_floor_dBm + self.sinr_dB

    @property
    def rx_noise_floor_dBm(self) -> float:
        return self.noise_figure_dB + self.thermal_noise_dBm

    def set_position(self, pos: Position) -> None:
        self.position = pos

    @property
    def num_subcarriers(self) -> int:
        return int(self.config['num_subcarriers'])

    @property
    def subcarrier_spacing_kHz(self) -> int:
        return int(self.config['subcarrier_spacing_kHz'])

    @property
    def rb_bandwidth_kHz(self) -> int:
        return self.num_subcarriers * self.subcarrier_spacing_kHz

    @property
    def max_tx_power_dBm(self) -> int:
        return self.config['max_tx_power_dBm']

    @property
    def antenna_height_m(self) -> float:
        return self.config['antenna_height_m']

    @property
    def tx_antenna_gain_dBi(self) -> float:
        return self.config['tx_antenna_gain_dBi']

    @property
    def rx_antenna_gain_dBi(self) -> float:
        return self.config['rx_antenna_gain_dBi']

    @property
    def noise_figure_dB(self) -> float:
        return self.config['noise_figure_dB']

    @property
    def thermal_noise_dBm(self) -> float:
        return self.config['thermal_noise_dBm']

    @property
    def sinr_dB(self) -> float:
        return self.config['sinr_dB']

    @property
    def ix_margin_dB(self) -> float:
        return self.config['ix_margin_dB']


class BaseStation(Device):
    def __init__(self, id_, config: dict = None) -> None:
        super().__init__(id_, merge_dicts(dict(DEFAULT_BASE_STATION_CONFIG), config or {}))

    def eirp_dBm(self, tx_pwr_dBm: float) -> float:
        return super().eirp_dBm(tx_pwr_dBm) - self.cable_loss_dB + self.masthead_amplifier_gain_dB

    def rx_signal_level_dBm(self, eirp_dBm: float, path_loss_dB: float) -> float:
        return super().rx_signal_level_dBm(eirp_dBm, path_loss_dB) \
               - self.cable_loss_dB \
               + self.masthead_amplifier_gain_dB

    @property
    def cable_loss_dB(self) -> float:
        return self.config['cable_loss_dB']

    @property
    def masthead_amplifier_gain_dB(self) -> float:
        return self.config['masthead_amplifier_gain_dB']

    def __repr__(self) -> str:
        return f'<BS:{self.id}>'


class UserEquipment(Device):
    def __init__(self, id_, config: dict = None) -> None:
        super().__init__(id_, merge_dicts(dict(DEFAULT_UE_CONFIG), config or {}))

    def eirp_dBm(self, tx_pwr_dBm: float) -> float:
        return super().eirp_dBm(tx_pwr_dBm) - self.body_loss_dB

    def rx_signal_level_dBm(self, eirp_dBm: float, path_loss_dB: float) -> float:
        return super().rx_signal_level_dBm(eirp_dBm, path_loss_dB) - self.body_loss_dB

    @property
    def control_channel_overhead_dB(self) -> float:
        return self.config['control_channel_overhead_dB']

    @property
    def body_loss_dB(self) -> float:
        return self.config['body_loss_dB']

    def __repr__(self) -> str:
        return f'<UE:{self.id}>'



from collections import defaultdict
from math import log2
from typing import Dict, Tuple

from .action import Action
from .channel import Channel
from .conversion import dB_to_linear, linear_to_dB
from .envs.devices import Devices
from .id import Id
from .path_loss import PathLoss
from .traffic_model import TrafficModel


class Simulator:
    def __init__(self, devices: Devices, traffic_model: TrafficModel, path_loss: PathLoss) -> None:
        super().__init__()
        self.devices: Devices = devices
        self.traffic_model: TrafficModel = traffic_model
        self.path_loss: PathLoss = path_loss
        self.channels: Dict[Tuple[Id, Id], Channel] = {}

    def reset(self):
        self.channels.clear()

    def step(self, actions: Dict[Id, Action]) -> dict:
        self._generate_traffic(actions)
        sinrs_db = self._calculate_sinrs()
        capacities = self._calculate_network_capacity(sinrs_db)

        return {
            'sinrs_db': sinrs_db,
            'snrs_db': self._calculate_snrs(),
            'rate_bps': self._calculate_rates(sinrs_db),
            'capacity_mbps': capacities,
        }

    def _generate_traffic(self, actions: Dict[Id, Action]) -> None:
        # automated traffic
        # self.channels = self.traffic_model.get_traffic()
        # supplied actions
        for action in actions.values():
            tx, rx = self.devices[action.tx_id], self.devices[action.rx_id]
            self.channels[(tx.id, rx.id)] = Channel(tx, rx, action.mode, action.rb, action.tx_pwr_dBm)

    def _calculate_sinrs(self) -> Dict[Tuple[Id, Id], float]:
        # group channels by RB
        rbs = defaultdict(set)
        for channel in self.channels.values():
            rbs[channel.rb].add(channel)

        sinrs_db = {}
        for (tx_id, rx_id), channel in self.channels.items():
            tx, rx = channel.tx, channel.rx
            rx_pwr_dBm = rx.rx_signal_level_dBm(tx.eirp_dBm(channel.tx_pwr_dBm), self.path_loss(tx, rx))

            ix_channels = rbs[channel.rb].difference({channel})
            sum_ix_pwr_mW = 0.0
            for ix_channel in ix_channels:
                ix_tx = ix_channel.tx
                ix_eirp_dBm = ix_tx.eirp_dBm(ix_channel.tx_pwr_dBm)
                ix_path_loss_dB = self.path_loss(ix_tx, rx)
                sum_ix_pwr_mW += dB_to_linear(ix_eirp_dBm - ix_path_loss_dB)

            # noise_mW = dB_to_linear(rx.thermal_noise_dBm)  # @todo this can be memoized
            # ix_and_noise_mW = sum_ix_pwr_mW + noise_mW
            # sinrs_db[(tx_id, rx_id)] = rx_pwr_dBm - linear_to_dB(ix_and_noise_mW)
            sinrs_db[(tx_id, rx_id)] = \
                float(rx_pwr_dBm - linear_to_dB(sum_ix_pwr_mW + dB_to_linear(rx.thermal_noise_dBm)))
        return sinrs_db

    def _calculate_snrs(self) -> Dict[Tuple[Id, Id], float]:
        SNRs_dB = {}
        for ids, channel in self.channels.items():
            tx, rx = channel.tx, channel.rx
            rx_pwr_dBm = rx.rx_signal_level_dBm(tx.eirp_dBm(channel.tx_pwr_dBm), self.path_loss(tx, rx))
            SNRs_dB[ids] = float(rx_pwr_dBm - rx.thermal_noise_dBm)
        return SNRs_dB

    def _calculate_rates(self, sinrs_db: Dict[Tuple[Id, Id], float]) -> Dict[Tuple[Id, Id], float]:
        rates_bps = {}
        for (tx_id, rx_id), sinr_db in sinrs_db.items():
            _, rx = self.devices[tx_id], self.devices[rx_id]
            # max_path_loss_dB = rx.max_path_loss_dB(tx.eirp_dBm())
            if sinr_db > rx.rx_sensitivity_dBm:
                rates_bps[(tx_id, rx_id)] = float(log2(1 + dB_to_linear(sinr_db)))
            else:
                rates_bps[(tx_id, rx_id)] = 0.0
        return rates_bps

    def _calculate_throughput_lte(self, sinrs_db: Dict[Tuple[Id, Id], float]) -> Dict[Tuple[Id, Id], float]:
        capacities = {}
        num_rbs = 100  # 100RBs @ 20MHz channel bandwidth
        num_re = 12 * 7 * 2  # num_subcarriers * num_symbols (short CP) * num_slots/subframe
        for (tx_id, rx_id), sinr_dB in sinrs_db.items():
            _, rx = self.devices[tx_id], self.devices[rx_id]
            if sinr_dB > rx.rx_sensitivity_dBm:
                num_bits = 6  # 64QAM
                capacity_b_per_ms = num_rbs * num_re * num_bits
                capacity_mbps = capacity_b_per_ms / 1000
                capacities[(tx_id, rx_id)] = capacity_mbps
            else:
                capacities[(tx_id, rx_id)] = 0
        return capacities

    def _calculate_network_capacity(self, sinrs_db: Dict[Tuple[Id, Id], float]) -> Dict[Tuple[Id, Id], float]:
        capacities_mbps = {}
        for (tx_id, rx_id), sinr_db in sinrs_db.items():
            tx, rx = self.devices[tx_id], self.devices[rx_id]
            # max_path_loss_dB = rx.max_path_loss_dB(tx.eirp_dBm())
            if sinr_db > rx.rx_sensitivity_dBm:
                b = tx.rb_bandwidth_kHz * 1000
                capacities_mbps[(tx_id, rx_id)] = float(1e-6 * b * log2(1 + dB_to_linear(sinr_db)))
            else:
                capacities_mbps[(tx_id, rx_id)] = 0.0
        return capacities_mbps



"""Examples of how to save and load an environment's device configuration from file.

To make experiments repeatable and compare algorithms, you may wish to fix UE positions.
This can be achieved by saving device configurations to an editable JSON file and
reloading the same configuration each experiment.

GymD2D uses Python's inbuilt pathlib paths to refer to files as it simplifies file handling.
"""

from pathlib import Path

import gym
import gym_d2d

# SAVING DEVICE CONFIG
# --------------------
# create a new environment and use `reset()` to generate random UE positions
env = gym.make('D2DEnv-v0')
env.reset()
# save the existing device configuration to a JSON file
env.save_device_config(Path.cwd() / 'device_config.json')


# LOADING DEVICE CONFIG
# ---------------------
# initialise a new environment with a Pathlib path to a device config file
env_config = {'device_config_file': Path.cwd() / 'device_config.json'}
env = gym.make('D2DEnv-v0', env_config=env_config)
# calls to `reset()` will use the device configuration from file---if the device ID is in it.
env.reset()



from abc import ABC, abstractmethod
from typing import Dict

from gym import Space, spaces
import numpy as np

from gym_d2d.id import Id
from gym_d2d.simulator import Simulator


class ObsFunction(ABC):
    def __init__(self, simulator: Simulator) -> None:
        super().__init__()
        self.simulator: Simulator = simulator

    @abstractmethod
    def get_obs_space(self, env_config: dict) -> Space:
        pass

    @abstractmethod
    def get_state(self, results: dict) -> Dict[Id, np.array]:
        pass


class LinearObsFunction(ObsFunction):
    def get_obs_space(self, env_config: dict) -> Space:
        r = env_config['cell_radius_m']
        num_txs = env_config['num_cues'] + env_config['num_due_pairs']
        num_obs = 6  # tx_x, tx_y, rx_x, rx_y, sinr, snr
        obs_shape = (num_obs * num_txs,)
        return spaces.Box(low=-r, high=r, shape=obs_shape)

    def get_state(self, results: dict) -> Dict[Id, np.array]:
        obses = {}
        for (tx_id, rx_id), channel in self.simulator.channels.items():
            obses[tx_id] = list(channel.tx.position.as_tuple() + channel.rx.position.as_tuple())
            obses[tx_id].append(results['sinrs_db'][(tx_id, rx_id)])
            obses[tx_id].append(results['snrs_db'][(tx_id, rx_id)])

        obs_dict = {}
        for tx_id in obses:
            tx_obs_copy = obses[tx_id][:]
            for other_tx_id, other_obs in obses.items():
                if other_tx_id != tx_id:
                    tx_obs_copy.extend(other_obs)
            obs_dict[tx_id] = np.array(tx_obs_copy)

        return obs_dict



from collections.abc import Mapping
from typing import Dict, Tuple

from gym_d2d.device import Device, BaseStation, UserEquipment
from gym_d2d.id import Id


class Devices(Mapping):
    def __init__(self,
                 bs: BaseStation,
                 cues: Dict[Id, UserEquipment],
                 due_pairs: Dict[Tuple[Id, Id], Tuple[UserEquipment, UserEquipment]]
                 ) -> None:
        super().__init__()
        self.bs: BaseStation = bs
        self.cues: Dict[Id, UserEquipment] = cues
        self.dues: Dict[Tuple[Id, Id], Tuple[UserEquipment, UserEquipment]] = due_pairs
        self.due_pairs = {}
        self.due_pairs_inv = {}
        self._devices: Dict[Id, Device] = {bs.id: bs, **cues}
        for (tx_id, rx_id), (tx, rx) in due_pairs.items():
            self._devices[tx_id] = tx
            self._devices[rx_id] = rx
            self.due_pairs[tx_id] = rx_id
            self.due_pairs_inv[rx_id] = tx_id

    def __getitem__(self, key):
        return self._devices[key]

    def __len__(self):
        return len(self._devices)

    def __iter__(self):
        return iter(self._devices)



from typing import Dict, List, Tuple

from .channel import Channel
from .device import BaseStation, UserEquipment
from .id import Id
from .link_type import LinkType


class TrafficModel:
    def __init__(self, bs: BaseStation, ues: List[UserEquipment], num_rbs: int) -> None:
        super().__init__()
        self.bs: BaseStation = bs
        self.ues: List[UserEquipment] = ues
        self.num_rbs: int = num_rbs

    def get_traffic(self) -> Dict[Tuple[Id, Id], Channel]:
        pass


class UplinkTrafficModel(TrafficModel):
    def get_traffic(self) -> Dict[Tuple[Id, Id], Channel]:
        rb = 0
        traffic = {}
        for ue in self.ues:
            traffic[(ue.id, self.bs.id)] = Channel(ue, self.bs, LinkType.UPLINK, rb, ue.max_tx_power_dBm)
            rb = (rb + 1) % self.num_rbs
        return traffic


class DownlinkTrafficModel(TrafficModel):
    def get_traffic(self) -> Dict[Tuple[Id, Id], Channel]:
        rb = 0
        traffic = {}
        for ue in self.ues:
            traffic[(self.bs.id, ue.id)] = Channel(self.bs, ue, LinkType.DOWNLINK, rb, ue.max_tx_power_dBm)
            rb = (rb + 1) % self.num_rbs
        return traffic



"""Example of how to visualise BS & UE locations.

Requires matplotlib to be installed separately (not included in package requirements).
"""

from pathlib import Path

import gym
import gym_d2d
from gym_d2d import plot_devices

# create a new environment, optionally load a device configuration from file,
# and use `reset()` to initialise the environment
env_config = {
    'num_cues': 15,
    'num_due_pairs': 15,
    'cell_radius_m': 500.0,
    'd2d_radius_m': 50.0,
    # 'device_config_file': Path.cwd() / 'device_config.json'
}
env = gym.make('D2DEnv-v0', env_config=env_config)
env.reset()

# use the plot helper to visualise device positions
plot_devices(env, 'plot.png')



from math import log10, pow


def dB_to_linear(dB: float) -> float:
    """Convert decibel values to linear scale.

    Args:
        dB: The value in decibels.

    Returns:
        The value in a linear scale.
    """
    return pow(10, dB / 10)


def linear_to_dB(linear: float) -> float:
    """Converts linear values to decibel scale.

    Args:
        linear: The linear value to convert.

    Returns:
        The value in a decibel scale.
    """
    return 10 * log10(linear)


def dBm_to_W(dBm: float) -> float:
    return dB_to_linear(dBm) / 1000


def W_to_dBm(linear: float) -> float:
    return linear_to_dB(linear * 1000)



from pathlib import Path
from pprint import pprint

import gym
import gym_d2d

env_config = {
    'num_rbs': 3,
    'num_cues': 3,
    'num_due_pairs': 3,
    'due_max_tx_power_dBm': 2,
    'device_config_file': Path.cwd() / 'device_config.json',
}
env = gym.make('D2DEnv-v0', env_config=env_config)

obses = env.reset()
game_over = False
for _ in range(3):
    actions = {}
    for agent_id, obs in obses.items():
        if agent_id.startswith('due'):
            actions[agent_id] = env.action_space['due'].sample()
        elif agent_id.startswith('cue'):
            actions[agent_id] = env.action_space['cue'].sample()
        else:
            actions[agent_id] = env.action_space['mbs'].sample()

    obses, rewards, game_over, info = env.step(actions)
    print(obses)
    pprint(info)
    print('\n\n')



from .device import Device
from .link_type import LinkType


class Channel:
    def __init__(self, tx: Device, rx: Device, link_type: LinkType, rb: int, tx_pwr_dBm: float) -> None:
        super().__init__()
        self.tx: Device = tx
        self.rx: Device = rx
        self.link_type: LinkType = link_type
        self.rb: int = rb
        self.tx_pwr_dBm: float = tx_pwr_dBm

    def __repr__(self) -> str:
        return f'<{self.tx.id}>:<{self.rx.id}>'



from dataclasses import dataclass
import json
from pathlib import Path
from typing import Optional, Type

from .obs_fn import ObsFunction, LinearObsFunction
from .reward_fn import RewardFunction, SystemCapacityRewardFunction
from gym_d2d.path_loss import PathLoss, LogDistancePathLoss
from gym_d2d.traffic_model import TrafficModel, UplinkTrafficModel


@dataclass
class EnvConfig:
    num_rbs: int = 25
    num_cues: int = 25
    num_due_pairs: int = 25
    cell_radius_m: float = 500.0
    d2d_radius_m: float = 20.0
    due_min_tx_power_dBm: int = 0
    due_max_tx_power_dBm: int = 20
    cue_max_tx_power_dBm: int = 23
    mbs_max_tx_power_dBm: int = 46
    path_loss_model: Type[PathLoss] = LogDistancePathLoss
    traffic_model: Type[TrafficModel] = UplinkTrafficModel
    carrier_freq_GHz: float = 2.1
    num_subcarriers: int = 12
    subcarrier_spacing_kHz: int = 15
    channel_bandwidth_MHz: float = 20.0
    obs_fn: Type[ObsFunction] = LinearObsFunction
    reward_fn: Type[RewardFunction] = SystemCapacityRewardFunction
    compressed_info: bool = False
    device_config_file: Optional[Path] = None

    def __post_init__(self):
        self.devices = self.load_device_config()

    def load_device_config(self) -> dict:
        if isinstance(self.device_config_file, Path):
            with self.device_config_file.open(mode='r') as fid:
                return json.load(fid)
        else:
            return {}



import json
from pathlib import Path
from typing import Dict, Tuple, Any

import gym
from gym import spaces
import numpy as np

from .devices import Devices
from .env_config import EnvConfig
from gym_d2d.action import Action
from gym_d2d.device import BaseStation, UserEquipment
from gym_d2d.id import Id
from gym_d2d.link_type import LinkType
from gym_d2d.position import Position, get_random_position, get_random_position_nearby
from gym_d2d.simulator import Simulator


BASE_STATION_ID = 'mbs'
EPISODE_LENGTH = 10


def create_devices(config: EnvConfig) -> Devices:
    """Initialise devices: BSs, CUEs & DUE pairs as per the env config.

    :returns: A dataclass containing BSs, CUEs, and DUE pairs.
    """

    base_cfg = {
        'num_subcarriers': config.num_subcarriers,
        'subcarrier_spacing_kHz': config.subcarrier_spacing_kHz,
    }

    # create macro base station
    cfg = config.devices[BASE_STATION_ID]['config'] if BASE_STATION_ID in config.devices else base_cfg
    bs = BaseStation(Id(BASE_STATION_ID), cfg)

    # create cellular UEs
    cues = {}
    default_cue_cfg = {**base_cfg, **{'max_tx_power_dBm': config.cue_max_tx_power_dBm}}
    for i in range(config.num_cues):
        cue_id = Id(f'cue{i:02d}')
        cfg = config.devices[cue_id]['config'] if cue_id in config.devices else default_cue_cfg
        cues[cue_id] = UserEquipment(cue_id, cfg)

    # create D2D UE pairs
    dues = {}
    due_cfg = {**base_cfg, **{'max_tx_power_dBm': config.due_max_tx_power_dBm}}
    for i in range(0, (config.num_due_pairs * 2), 2):
        due_tx_id, due_rx_id = Id(f'due{i:02d}'), Id(f'due{i + 1:02d}')

        due_tx_cfg = config.devices[due_tx_id]['config'] if due_tx_id in config.devices else due_cfg
        due_tx = UserEquipment(due_tx_id, due_tx_cfg)

        due_rx_cfg = config.devices[due_rx_id]['config'] if due_rx_id in config.devices else due_cfg
        due_rx = UserEquipment(due_rx_id, due_rx_cfg)

        dues[(due_tx.id, due_rx.id)] = due_tx, due_rx

    return Devices(bs, cues, dues)


class D2DEnv(gym.Env):
    metadata = {'render.modes': ['human']}

    def __init__(self, env_config=None) -> None:
        super().__init__()
        self.config = EnvConfig(**env_config or {})
        devices = create_devices(self.config)
        self.simulator = Simulator(
            devices,
            self.config.traffic_model(devices.bs, list(devices.cues.values()), self.config.num_rbs),
            self.config.path_loss_model(self.config.carrier_freq_GHz)
        )
        self.obs_fn = self.config.obs_fn(self.simulator)
        self.observation_space = self.obs_fn.get_obs_space(self.config.__dict__)
        self.num_pwr_actions = {  # +1 because include max value, i.e. from [0, ..., max]
            'due': self.config.due_max_tx_power_dBm - self.config.due_min_tx_power_dBm + 1,
            'cue': self.config.cue_max_tx_power_dBm + 1,
            'mbs': self.config.mbs_max_tx_power_dBm + 1
        }
        self.action_space = spaces.Dict({
            'due': spaces.Discrete(self.config.num_rbs * self.num_pwr_actions['due']),
            'cue': spaces.Discrete(self.config.num_rbs * self.num_pwr_actions['cue']),
            'mbs': spaces.Discrete(self.config.num_rbs * self.num_pwr_actions['mbs']),
        })
        self.reward_fn = self.config.reward_fn(self.simulator)
        self.num_steps = 0

    def reset(self):
        self.num_steps = 0
        for device in self.simulator.devices.values():
            if device.id == BASE_STATION_ID:
                pos = Position(0, 0)  # assume MBS fixed at (0,0) and everything else builds around it
            elif device.id in self.config.devices:
                pos = Position(*self.config.devices[device.id]['position'])
            elif any(device.id in d for d in [self.simulator.devices.cues, self.simulator.devices.due_pairs]):
                pos = get_random_position(self.config.cell_radius_m)
            elif device.id in self.simulator.devices.due_pairs_inv:
                due_tx_id = self.simulator.devices.due_pairs_inv[device.id]
                due_tx = self.simulator.devices[due_tx_id]
                pos = get_random_position_nearby(self.config.cell_radius_m, due_tx.position, self.config.d2d_radius_m)
            else:
                raise ValueError(f'Invalid configuration for device "{device.id}".')
            device.set_position(pos)

        self.simulator.reset()
        # take a step with random D2D actions to generate initial SINRs
        random_actions = self._reset_random_actions()
        results = self.simulator.step(random_actions)
        obs = self.obs_fn.get_state(results)
        return obs

    def _reset_random_actions(self) -> Dict[Id, Action]:
        cue_actions = {_id: self._extract_action(_id, self.action_space['cue'].sample())
                       for _id in self.simulator.devices.cues.keys()}
        due_actions = {_id: self._extract_action(_id, self.action_space['due'].sample())
                       for _id, _ in self.simulator.devices.dues.keys()}
        return {**cue_actions, **due_actions}

    def step(self, actions):
        actions = self._extract_actions(actions)
        results = self.simulator.step(actions)
        self.num_steps += 1
        obs = self.obs_fn.get_state(results)
        rewards = self.reward_fn(results)
        game_over = {'__all__': self.num_steps >= EPISODE_LENGTH}
        info = self._info(actions, results)

        return obs, rewards, game_over, info

    def _extract_actions(self, actions: Dict[Id, object]) -> Dict[Id, Action]:
        return {tx_id: self._extract_action(tx_id, action) for tx_id, action in actions.items()}

    def _extract_action(self, tx_id: Id, action) -> Action:
        if tx_id in self.simulator.devices.due_pairs:
            rx_id = self.simulator.devices.due_pairs[tx_id]
            link_type = LinkType.SIDELINK
            rb, tx_pwr_dBm = self._decode_action(action, 'due')
        elif tx_id in self.simulator.devices.cues:
            rx_id = self.simulator.devices.bs.id
            link_type = LinkType.UPLINK
            rb, tx_pwr_dBm = self._decode_action(action, 'cue')
        else:
            rx_id = self.simulator.devices.cues[tx_id]
            link_type = LinkType.DOWNLINK
            rb, tx_pwr_dBm = self._decode_action(action, 'mbs')
        return Action(tx_id, rx_id, link_type, rb, tx_pwr_dBm)

    def _decode_action(self, action: Any, tx_type: str) -> Tuple[int, int]:
        if isinstance(action, (int, np.integer)):
            rb = action // self.num_pwr_actions[tx_type]
            tx_pwr_dBm = action % self.num_pwr_actions[tx_type]
        elif isinstance(action, np.ndarray) and action.ndim == 2:
            rb, tx_pwr_dBm = action
        else:
            raise ValueError(f'Unable to decode action type "{type(action)}"')
        return int(rb), int(tx_pwr_dBm)

    def _info(self, actions: Dict[Id, Action], results: dict):
        return {
            tx_id: {
                'rb': actions[tx_id].rb,
                'tx_pwr_dbm': actions[tx_id].tx_pwr_dBm,
                # 'channel_gains_db': results['channel_gains_db'][(tx_id, rx_id)],
                'snr_db': results['snrs_db'][(tx_id, rx_id)],
                'sinr_db': sinr_db,
                'rate_bps': results['rate_bps'][(tx_id, rx_id)],
                'capacity_mbps': results['capacity_mbps'][(tx_id, rx_id)],
            } for (tx_id, rx_id), sinr_db in results['sinrs_db'].items()}

    def render(self, mode='human'):
        obs = self.obs_fn.get_state({})  # @todo need to find a way to handle SINRs here
        print(obs)

    def save_device_config(self, config_file: Path) -> None:
        """Save the environment's device configuration in a JSON file.

        :param config_file: The filepath to save to.
        """
        config = {device.id: {
                'position': device.position.as_tuple(),
                'config': device.config,
            } for device in self.simulator.devices.values()}
        with config_file.open(mode='w') as fid:
            json.dump(config, fid)



def merge_dicts(original: dict, other: dict) -> dict:
    """Merge two dicts (in place), overwriting values in original.

    :param original: The original dict values being used.
    :param other: The dict to overwrite with.
    :return: The overwritten dict.
    """
    for k, v in other.items():
        if k in original:
            if isinstance(original[k], dict) and isinstance(other[k], dict):
                merge_dicts(original[k], other[k])
            else:
                original[k] = other[k]  # overwrite
        else:
            original[k] = other[k]
    return original


def plot_devices(env, out_file: str = ''):
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError('`plot_devices()` requires matplotlib')

    cue_xs, cue_ys = [], []
    for cue_id, cue in env.simulator.devices.cues.items():
        cue_xs.append(cue.position.x)
        cue_ys.append(cue.position.y)
    due_tx_xs, due_tx_ys = [], []
    due_rx_xs, due_rx_ys = [], []
    for due_tx_id, due_rx_id in env.simulator.devices.due_pairs.items():
        due_tx = env.simulator.devices[due_tx_id]
        due_tx_xs.append(due_tx.position.x)
        due_tx_ys.append(due_tx.position.y)
        due_rx = env.simulator.devices[due_rx_id]
        due_rx_xs.append(due_rx.position.x)
        due_rx_ys.append(due_rx.position.y)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    mbs_pos = env.simulator.devices.bs.position
    ax.add_artist(plt.Circle(mbs_pos.as_tuple(), env.config.cell_radius_m, color='b', alpha=0.1))
    ax.scatter(due_tx_xs, due_tx_ys, c='r', label='DUE_TX')
    ax.scatter(due_rx_xs, due_rx_ys, c='m', label='DUE_RX')
    ax.scatter(cue_xs, cue_ys, c='b', label='CUE')
    ax.scatter(mbs_pos.x, mbs_pos.y, c='k', label='MBS')
    ax.legend()
    if out_file:
        plt.savefig(out_file)
    plt.show()



from math import log10

import gym
from gym_d2d.device import Device
from gym_d2d.path_loss import PathLoss


class FooPathLoss(PathLoss):
    """Define your own custom path loss implementation.
    If you have a path loss model that isn't included, why not submit a pull request so others can use it too.
    """

    def __call__(self, tx: Device, rx: Device) -> float:
        d = tx.position.distance(rx.position)
        foo_path_loss = 20 * log10(d) - tx.tx_antenna_gain_dBi - rx.rx_antenna_gain_dBi
        return foo_path_loss


# pass your custom path loss model class (not instance) in the env_config
# alternatively, you can import and use one of the inbuilt path loss models
env_config = {'path_loss_model': FooPathLoss}
env = gym.make('D2DEnv-v0', env_config=env_config)

# run the standard agent loop, no change here
agent = 'D2DAgent()'
obses = env.reset()
game_over = False
for _ in range(10):
    actions = {}
    for agent_id, obs in obses.items():
        actions[agent_id] = env.action_space['due'].sample()

    obses, rewards, game_over, info = env.step(actions)
    print(obses)



from setuptools import setup, find_packages


setup(
    name='gym_d2d',
    version='0.0.3',
    description='Device-to-Device (D2D) communication OpenAI Gym environment',
    keywords='open ai gym environment rl agent d2d cellular offload resource allocation',
    url='https://github.com/davidcotton/gym-d2d',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    install_requires=['gym>=0.9.6', 'numpy'],
    extras_require={
        'dev': ['flake8', 'pytest', 'pytest-cov', 'pytest-sugar']
    },
    clasifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



from abc import ABC, abstractmethod
from enum import Enum
from math import log10, pi
from random import gauss

from .device import Device


SPEED_OF_LIGHT = 299792458  # m/s


class PathLoss(ABC):
    def __init__(self, carrier_freq_GHz: float) -> None:
        super().__init__()
        self.carrier_freq_GHz: float = carrier_freq_GHz

    @abstractmethod
    def __call__(self, tx: Device, rx: Device) -> float:
        """Calculate the path loss between a transmitter and a receiver.

        :param tx: The transmitting device.
        :param rx: The receiving device.
        :return: The free space path loss in dB.
        """
        pass


def pl_constant_dB(carrier_freq_GHz: float, ple: float) -> float:
    """Calculate the constant part of Log-Distance Path Loss equation.

    We assume a fixed carrier frequency for all communications in our simulation.
    This means that only the distance and antenna gain parts of the Log-Distance PL equation will be changing,
    so we can memoize the freq + speed of light part to save computation.

    :param carrier_freq_GHz: The carrier frequencies in Ghz.
    :param ple: The path loss exponent.
    :return: The log-distance path loss constant in dB.
    """
    return 10 * ple * log10(carrier_freq_GHz * 1e9) + 10 * ple * log10((4 * pi) / SPEED_OF_LIGHT)


class LogDistancePathLoss(PathLoss):
    def __init__(self, carrier_freq_GHz: float, ple: float = 2.0) -> None:
        super().__init__(carrier_freq_GHz)
        self.ple: float = ple  # path loss exponent (2.0 in free space, 3.5 in crowded env)
        self.pl_constant_dB = pl_constant_dB(carrier_freq_GHz, ple)

    def __call__(self, tx: Device, rx: Device) -> float:
        """Calculate the loss of signal strength in free space.

        LDPL = 10nlog_10(d) + 10nlog_10(f) + 10nlog_10(4pi/c)

        Where:
            d: distance in metres
            f: the carrier frequency in Hz
            c: speed of light (m/s)

        :param tx: The transmitting device.
        :param rx: The receiving device.
        :return: The log-distance path loss in dB.
        """

        return self._log_distance_path_loss(tx.position.distance(rx.position))

    def _log_distance_path_loss(self, dist_m: float) -> float:
        return 10 * self.ple * log10(dist_m) + self.pl_constant_dB


class ShadowingPathLoss(LogDistancePathLoss):
    def __init__(self, carrier_freq_GHz: float, ple: float = 2.0, d0_m: float = 100.0, chi_dB: float = 2.7) -> None:
        super().__init__(carrier_freq_GHz, ple)
        self.d0_m: float = d0_m  # shadowing close-in reference distance (metres)
        self.chi_dB: float = chi_dB  # shadowing standard deviation (dB), typically 2.7 to 3.5

    def __call__(self, tx: Device, rx: Device) -> float:
        d = tx.position.distance(rx.position)
        assert self.d0_m < d
        ldpl = self._log_distance_path_loss(self.d0_m)
        return ldpl + 10 * self.ple * log10(d / self.d0_m) + gauss(0, self.chi_dB)


class AreaType(Enum):
    RURAL = 0
    SUBURBAN = 1
    URBAN = 2


class CostHataPathLoss(PathLoss):
    def __init__(self, carrier_freq_GHz: float, area_type: AreaType = AreaType.SUBURBAN) -> None:
        super().__init__(carrier_freq_GHz)
        self.area_type: AreaType = area_type

    def __call__(self, tx: Device, rx: Device) -> float:
        """
        Lb = 46.3 + 33.9log_10(f) - 13.82log_10(h_tx) - a(h_rx,f) + (44.9 - 6.55log_10(h_tx)log_10(d) + C
        """
        f = self.carrier_freq_GHz * 1000  # transmission freq (MHz)
        d = tx.position.distance(rx.position) / 1000  # link distance (Km)
        h_tx = tx.antenna_height_m  # TX antenna height (m)
        h_rx = rx.antenna_height_m  # RX antenna height (m)
        a_hc = self._ms_h_correction(f, h_rx)  # MS antenna height correction factor
        # constant offset (dB), 0 for med cities & suburban, 3 for metropolitan areas
        c = 3 if self.area_type == AreaType.URBAN else 0
        pl = 46.3 + 33.9 * log10(f) - 13.82 * log10(h_tx) - a_hc + (44.9 - 6.55 * log10(h_tx)) * log10(d) + c
        return pl

    def _ms_h_correction(self, f: float, h_rx: float) -> float:
        """RX antenna height correction factor.

        :param f: Carrier freq (MHz)
        :param h_rx: Receiving station antenna height (m)
        :return: The height correction factor.
        """
        if self.area_type == AreaType.URBAN:
            if f >= 200:
                a_hc = 8.29 * (log10(1.54 * h_rx)) ** 2 - 1.1
            else:
                a_hc = 3.2 * (log10(11.75 * h_rx)) ** 2 - 4.97
        else:
            a_hc = (1.1 * log10(f) - 0.7) * h_rx - (1.56 * log10(f) - 0.8)
        return a_hc



from abc import ABC, abstractmethod
from collections import defaultdict
from math import log2
from typing import Dict

from gym_d2d.id import Id
from gym_d2d.simulator import Simulator
from gym_d2d.conversion import dB_to_linear
from gym_d2d.link_type import LinkType


class RewardFunction(ABC):
    def __init__(self, simulator: Simulator) -> None:
        super().__init__()
        self.simulator: Simulator = simulator

    @abstractmethod
    def __call__(self, results: dict) -> Dict[Id, float]:
        pass


class SystemCapacityRewardFunction(RewardFunction):
    def __call__(self, results: dict) -> Dict[Id, float]:
        # group by RB
        rbs = defaultdict(set)
        for ids, channel in self.simulator.channels.items():
            rbs[channel.rb].add(ids)

        reward = -1
        _break = False
        for tx_id, rx_id in self.simulator.devices.due_pairs.items():
            if _break:
                break
            rb = self.simulator.channels[(tx_id, rx_id)].rb
            ix_channels = rbs[rb].difference({(tx_id, rx_id)})
            for ix_tx_id, ix_rx_id in ix_channels:
                if ix_tx_id in self.simulator.devices.due_pairs:
                    continue
                if results['capacity_mbps'][(ix_tx_id, ix_rx_id)] <= 0:
                    _break = True
                    break
        else:
            sum_capacity = sum(results['capacity_mbps'].values())
            reward = sum_capacity / len(self.simulator.devices.due_pairs)

        return {tx_id: reward for tx_id in self.simulator.devices.due_pairs.keys()}


class DueShannonRewardFunction(RewardFunction):
    def __init__(self, simulator: Simulator) -> None:
        super().__init__(simulator)
        self.min_sinr = -70.0

    def __call__(self, results: dict) -> Dict[Id, float]:
        rewards = {}
        for tx_id, rx_id in self.simulator.devices.due_pairs.items():
            sinr = results['sinrs_db'][(tx_id, rx_id)]
            if sinr >= self.min_sinr:
                rewards[tx_id] = log2(1 + dB_to_linear(sinr))
            else:
                rewards[tx_id] = -1
        return rewards


class CueSinrShannonRewardFunction(RewardFunction):
    def __init__(self, simulator: Simulator, sinr_threshold_dB=0.0) -> None:
        super().__init__(simulator)
        self.sinr_threshold_dB: float = float(sinr_threshold_dB)

    def __call__(self, results: dict) -> Dict[Id, float]:
        # group channels by RB
        rbs = defaultdict(set)
        for ids, channel in self.simulator.channels.items():
            rbs[channel.rb].add(channel)

        rewards = {}
        for tx_id, rx_id in self.simulator.devices.due_pairs.items():
            channel = self.simulator.channels[(tx_id, rx_id)]
            ix_channels = rbs[channel.rb].difference({channel})
            rewards[tx_id] = -1
            for ix_channel in ix_channels:
                if ix_channel.link_type != LinkType.SIDELINK:
                    cue_sinr_dB = results['sinrs_db'][(ix_channel.tx.id, ix_channel.rx.id)]
                    if cue_sinr_dB < self.sinr_threshold_dB:
                        break
            else:
                rewards[tx_id] = log2(1 + dB_to_linear(results['sinrs_db'][(tx_id, rx_id)]))
        return rewards
